from collections.abc import Mapping
from typing import Iterable

import pydantic as pd


class BaseModel(pd.BaseModel, Mapping):
    def __iter__(self) -> Iterable[str]:  # type: ignore[override]
        return iter([field_name for field_name in self.model_fields_set])

    def __len__(self):
        return len(self.model_fields)

    def __getitem__(self, key):
        try:
            return getattr(self, key)
        except AttributeError:
            raise KeyError(key)



import inspect
from functools import partial
from typing import (
    Any,
    ClassVar,
    Iterable,
    Literal,
    LiteralString,
    Self,
    TypeGuard,
    TypeVar,
    get_origin,
)

from pydantic import GetCoreSchemaHandler
from pydantic_core import core_schema

from fhir_slicing.slice import BaseSlice, slice_validator
from fhir_slicing.slice_schema import get_slice_union_schema
from fhir_slicing.typing import ElementArray

from .utils import FHIRType, get_source_type

TUrl = TypeVar("TUrl", bound=LiteralString)
TFhirType = TypeVar("TFhirType", bound=FHIRType)
TPythonType = TypeVar("TPythonType")


# resource.extension["itemControl"].valueCodeableConcept.coding["tiro"].code

DiscriminatorType = Literal["value", "exists", "type"]
TDefaultElement = TypeVar("TDefaultElement")

FIELD_NAME_TO_SLICE_NAME = {"_": "@default"}


def get_all_parent_classes(cls):
    return inspect.getmro(cls)


def get_slice_annotations(element_array_cls: type[Any]) -> dict[str, type]:
    return {
        FIELD_NAME_TO_SLICE_NAME.get(field_name, field_name): annotation
        for cls in get_all_parent_classes(element_array_cls)
        for field_name, annotation in inspect.get_annotations(cls).items()
        if get_origin(annotation) is not ClassVar
    }


class BaseElementArray[TElement](ElementArray[TElement]):
    """A collection of elements that can be sliced and named using a discriminator."""

    @classmethod
    def filter_elements_for_slice(cls, elements: Self, slice_name: str) -> Iterable[TElement]:
        """Get the slice name for a given element."""
        for element in elements:
            if cls.is_element_part_of_slice(element, slice_name):
                yield element

    @classmethod
    def is_element_part_of_slice(cls, element: TElement, slice_name: str) -> TypeGuard[TElement]:
        """Check if an element is part of a slice."""
        annotation = get_slice_annotations(cls)[slice_name]
        for element_type in get_source_type(annotation):
            if isinstance(element, element_type):
                return True
        return False

    @classmethod
    def get_validators(cls):
        def get_slice_validator(cls, slice_name):
            descriptor = inspect.getattr_static(cls, slice_name)
            if not isinstance(descriptor, BaseSlice):
                raise TypeError(f"Expected Slice, OptionalSlice or SliceList, got {type(descriptor)}")

        for field_name in inspect.get_annotations(cls).keys():
            descriptor = inspect.getattr_static(cls, field_name)
            if not isinstance(descriptor, BaseSlice):
                raise TypeError(f"Expected Slice, OptionalSlice or SliceList, got {type(descriptor)}")
            slice_name = FIELD_NAME_TO_SLICE_NAME.get(field_name, field_name)
            yield partial(slice_validator, slice_name=slice_name, slice_descriptor=descriptor)

    @classmethod
    def __get_pydantic_core_schema__(cls, source_type: Any, handler: GetCoreSchemaHandler):
        """Get the Pydantic core schema for the element array."""
        slice_union_schema = get_slice_union_schema(source_type, handler, slice_annotations=get_slice_annotations(cls))
        list_schema = core_schema.list_schema(slice_union_schema)
        # TODO add after validators for cardinality of each slice

        schema = list_schema
        for validator in cls.get_validators():
            schema = core_schema.no_info_after_validator_function(
                validator,
                schema,
            )
        return core_schema.json_or_python_schema(
            core_schema.no_info_after_validator_function(
                cls,
                schema,
            ),
            core_schema.union_schema(
                [
                    core_schema.is_instance_schema(cls),
                    core_schema.no_info_after_validator_function(cls, schema),
                ]
            ),
        )


if __name__ == "__main__":
    pass



from typing import Annotated, List, Optional

import pytest

from fhir_slicing.utils import (
    get_source_type,
)


@pytest.mark.parametrize(
    "source, target",
    [
        (Optional[str], (str,)),
        (List[str], (str,)),
        (str, (str,)),
        (List[str] | None, (str,)),
        (Annotated[List[str], "some annotation"], (str,)),
        (Annotated[List[str] | None, "some annotation"], (str,)),
        (Annotated[List[str | int] | None, "some annotation"], (str, int)),
    ],
)
def test_get_source_type(source, target):
    source_types = tuple(get_source_type(source))
    assert source_types == target


class Base:
    pass


class A(Base):
    pass


class B(Base):
    pass


class C:
    pass


@pytest.mark.parametrize(
    "source, expected, raises",
    [
        [A, Base, None],
        [B, Base, None],
        [C, Base, TypeError],
        [A | None, Base, None],
        [B | None, Base, None],
        [C | None, Base, TypeError],
        [Optional[A], Base, None],
        [Optional[B], Base, None],
        [Optional[C], Base, TypeError],
        [List[A], Base, None],
        [List[B], Base, None],
        [List[C], Base, TypeError],
        [List[A | B], Base, None],
        [List[B | C], Base, TypeError],
        [List[A | B] | None, Base, None],
        [List[A | B] | None | None, Base, None],
        [Annotated[List[A | B] | None, "some annotation"], Base, None],
        [Annotated[List[A | C] | None, "some annotation"], Base, TypeError],
    ],
)
def test_get_source_type_with_expect(source, expected, raises):
    source_types_iter = get_source_type(source, expect=expected)
    if raises:
        with pytest.raises(raises):
            tuple(source_types_iter)
    else:
        tuple(source_types_iter)



from typing import Annotated, Iterable, List, Optional, Self, TypeGuard, TypeVar

from pydantic import GetJsonSchemaHandler
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import core_schema


def is_null_schema(schema: JsonSchemaValue) -> bool:
    return "type" in schema and schema["type"] == "null"


class NotNullable:
    @classmethod
    def __get_pydantic_json_schema__(
        cls, core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        json_schema = handler(core_schema)
        assert "anyOf" in json_schema
        not_null_schema = next(
            (schema for schema in filter(lambda x: not is_null_schema(x), json_schema["anyOf"])), None
        )
        assert not_null_schema is not None
        return not_null_schema


T = TypeVar("T")
Omissible = Annotated[Optional[T], NotNullable()]


class ElementArray[TElement](List[TElement]):
    @classmethod
    def filter_elements_for_slice(cls, elements: Self, slice_name: str) -> Iterable[TElement]:
        """Get the slice name for a given element."""
        ...

    @classmethod
    def is_element_part_of_slice(cls, element: TElement, slice_name: str) -> TypeGuard[TElement]:
        """Check if an element is part of a slice."""
        ...



from typing import Literal

from pydantic import BaseModel

from fhir_slicing import OptionalSlice, SliceList, slice
from fhir_slicing.element_array import BaseElementArray
from fhir_slicing.usage_context import BaseUsageContext, UsageCoding

UsageContextType = Literal["http://terminology.hl7.org/CodeSystem/usage-context-type"]
AgeContextCoding = UsageCoding[UsageContextType, Literal["age"], None]
ProgramContextCoding = UsageCoding[UsageContextType, Literal["program"], None]


class Age(BaseModel):
    value: int
    system: Literal["http://unitsofmeasure.org"] = "http://unitsofmeasure.org"
    code: Literal["a"] = "a"
    unit: Literal["year"] = "year"


class AgeRange(BaseModel):
    low: Age
    high: Age


class AgeContext(BaseUsageContext[AgeContextCoding]):
    valueRange: AgeRange

    @classmethod
    def from_age_range(cls, low: int, high: int):
        return cls(
            valueRange=AgeRange(low=Age(value=low), high=Age(value=high)),
            code=AgeContextCoding(system="http://terminology.hl7.org/CodeSystem/usage-context-type", code="age"),
        )


class Reference(BaseModel):
    reference: str


class ProjectContext(BaseUsageContext[ProgramContextCoding]):
    valueReference: Reference

    @classmethod
    def from_reference(cls, reference: str):
        return cls(
            valueReference=Reference(reference=reference),
            code=ProgramContextCoding(
                system="http://terminology.hl7.org/CodeSystem/usage-context-type", code="program"
            ),
        )


def test_usage_context_array_from_usage_context_list():
    class UsageContextArray(BaseElementArray):
        project: SliceList[ProjectContext] = slice(0, "*")
        age_range: OptionalSlice[AgeContext] = slice(0, 1)
        _: SliceList[BaseUsageContext] = slice(0, "*")

    ctx_list = [
        ProjectContext.from_reference("https://example.com/project1"),
        AgeContext.from_age_range(18, 65),
        ProjectContext.from_reference("https://example.com/project2"),
        BaseUsageContext.model_validate(
            {
                "valueCodeableConcept": {
                    "coding": [{"system": "http://snomed.info/sct", "code": "123456789", "display": "test"}],
                },
                "code": {
                    "system": "http://terminology.hl7.org/CodeSystem/usage-context-type",
                    "code": "gender",
                    "display": "Gender",
                },
            }
        ),
    ]

    ctx_array = UsageContextArray(ctx_list)
    assert ctx_array.project == [ctx_list[0], ctx_list[2]]
    assert ctx_array.age_range == ctx_list[1]
    assert list(ctx_array) == ctx_list



from typing import Literal

import pytest
from pydantic import TypeAdapter
from pydantic_core import ValidationError

from fhir_slicing.base import BaseModel
from fhir_slicing.coding import BaseCoding, GeneralCoding, LOINCCoding, SCTCoding
from fhir_slicing.element_array import BaseElementArray
from fhir_slicing.slice import OptionalSlice, Slice, SliceList, slice


def test_multi_coding_concepts():
    class CodingArray(BaseElementArray):
        sct: Slice[SCTCoding] = slice(1, 1)
        loinc: OptionalSlice[LOINCCoding] = slice(0, 1)
        _: SliceList[GeneralCoding] = slice(0, "*")

    class CodeableConcept(BaseModel):
        coding: CodingArray
        text: str | None = None

    raw_concept = {
        "coding": [
            {"system": "http://snomed.info/sct", "code": "123456", "display": "Test"},
            {"system": "http://loinc.org", "code": "123456", "display": "Test"},
            {"system": "http://other.org", "code": "123456", "display": "Test"},
        ],
        "text": "Test",
    }

    concept = CodeableConcept.model_validate(raw_concept)

    assert concept.coding.sct.system == "http://snomed.info/sct"
    assert concept.coding.loinc is not None, "Expected loinc to be present"
    assert concept.coding.loinc.system == "http://loinc.org"

    assert (
        concept.model_dump(by_alias=True, exclude_none=True) == raw_concept
    ), "Expected model_dump to match raw_concept"


def test_coding_cardinality_max_items():
    class CodingArray(BaseElementArray):
        sct: Slice[SCTCoding] = slice(1, 1)
        loinc: SliceList[LOINCCoding] = slice(1, 10)
        _: SliceList[GeneralCoding] = slice(0, "*")

    class CodeableConcept(BaseModel):
        coding: CodingArray
        text: str | None = None

    raw_concept = {
        "coding": [
            {"system": "http://loinc.org", "code": "123456", "display": "Test"},
            {"system": "http://snomed.info/sct", "code": "123456", "display": "Test"},
            {"system": "http://snomed.info/sct", "code": "987654", "display": "Test2"},
            {"system": "http://other.org", "code": "123456", "display": "Test"},
        ],
        "text": "Test",
    }
    with pytest.raises(ValidationError) as exc_info:
        CodeableConcept.model_validate(raw_concept)
    error_msgs = [error["msg"] for error in exc_info.value.errors()]
    assert "Value error, Too many items in slice 'sct'" in error_msgs


def test_coding_cardinality_mmin_items():
    class CodingArray(BaseElementArray):
        sct: Slice[SCTCoding] = slice(1, 1)
        loinc: SliceList[LOINCCoding] = slice(2, 10)
        _: SliceList[GeneralCoding] = slice(0, "*")

    class CodeableConcept(BaseModel):
        coding: CodingArray
        text: str | None = None

    raw_concept = {
        "coding": [
            {"system": "http://loinc.org", "code": "123456", "display": "Test"},
            {"system": "http://snomed.info/sct", "code": "987654", "display": "Test2"},
            {"system": "http://other.org", "code": "123456", "display": "Test"},
        ],
        "text": "Test",
    }
    with pytest.raises(ValidationError) as exc_info:
        CodeableConcept.model_validate(raw_concept)
    error_msgs = [error["msg"] for error in exc_info.value.errors()]
    assert "Value error, Not enough items in slice 'loinc'" in error_msgs


def test_task_code():
    class AtticusTaskType(BaseCoding):
        code: Literal["complete-questionnaire", "process-response"]
        system: Literal["https://tiro.health/fhir/CodeSystem/atticus-task-type"]

    class TaskCodingArray(BaseElementArray[GeneralCoding | AtticusTaskType]):
        atticus_task_type: Slice[AtticusTaskType] = slice(1, 1)
        _: SliceList[GeneralCoding] = slice(0, "*")

    coding_array = TaskCodingArray(
        [AtticusTaskType(code="complete-questionnaire", system="https://tiro.health/fhir/CodeSystem/atticus-task-type")]
    )

    coding_array_json = TypeAdapter[TaskCodingArray](TaskCodingArray).dump_python(
        coding_array, by_alias=True, exclude_none=True
    )

    assert coding_array_json == [
        {"system": "https://tiro.health/fhir/CodeSystem/atticus-task-type", "code": "complete-questionnaire"}
    ]



from .base import BaseModel
from .coding import BaseCoding


class UsageCoding[TSystem: str, TCode: str, TVersion: str | None](BaseCoding[TSystem]):
    code: TCode
    version: TVersion | None = None

    def __eq__(self, other: object):
        if not isinstance(other, UsageCoding):
            return False
        return hash((self.system, self.code, self.version)) == hash((other.system, other.code, other.version))


class BaseUsageContext[TUsageCoding: UsageCoding](BaseModel):
    code: TUsageCoding
    model_config = {"extra": "allow"}



from typing import Literal, Protocol

from fhir_slicing.typing import Omissible

from .base import BaseModel


class BaseCoding[TSystem: str](BaseModel):
    system: TSystem


class GeneralCoding(BaseCoding):
    display: Omissible[str] = None
    code: str
    model_config = {"extra": "allow"}


class CodingProtocol[TSystem: str](Protocol):
    @property
    def system(self) -> TSystem: ...

    @property
    def code(self) -> str: ...

    @classmethod
    def get_system(cls) -> TSystem | None: ...


class SCTCoding(BaseCoding):
    system: Literal["http://snomed.info/sct"] = "http://snomed.info/sct"
    code: str
    display: str


class LOINCCoding(BaseCoding):
    system: Literal["http://loinc.org"] = "http://loinc.org"
    code: str
    display: str



from typing import Annotated, Literal

from pydantic import AfterValidator

from fhir_slicing import Slice, SliceList, slice
from fhir_slicing.base import BaseModel
from fhir_slicing.coding import GeneralCoding, LOINCCoding
from fhir_slicing.element_array import BaseElementArray


def test_blood_pressure_use_case():
    class Quantity(BaseModel):
        value: float
        unit: str

    class CodingArray(BaseElementArray[LOINCCoding | GeneralCoding]):
        loinc: Slice[LOINCCoding] = slice(1, 1)
        _: SliceList[GeneralCoding] = slice(0, "*")

    class CodeableConcept(BaseModel):
        coding: CodingArray
        text: str | None = None

    def is_systolic(value: CodeableConcept):
        if value.coding.loinc.code != "8480-6":
            raise ValueError("Not a systolic blood pressure")
        return value

    def is_diastolic(value: CodeableConcept):
        if value.coding.loinc.code != "8462-4":
            raise ValueError("Not a diastolic blood pressure")
        return value

    class SystolicPressureComponent(BaseModel):
        valueQuantity: Quantity
        code: Annotated[CodeableConcept, AfterValidator(is_systolic)]

    class DiastolicPressureComponent(BaseModel):
        valueQuantity: Quantity
        code: Annotated[CodeableConcept, AfterValidator(is_diastolic)]

    class BloodPressureComponents(BaseElementArray):
        systolic: Slice[SystolicPressureComponent] = slice(1, 1)
        diastolic: Slice[DiastolicPressureComponent] = slice(1, 1)

    class BloodPressure(BaseModel):
        resourceType: Literal["Observation"] = "Observation"
        code: CodeableConcept
        components: BloodPressureComponents

        @property
        def systolic(self):
            return self.components.systolic.valueQuantity.value

        @property
        def diastolic(self):
            return self.components.diastolic.valueQuantity.value

    blood_pressure = BloodPressure.model_validate(
        {
            "resourceType": "Observation",
            "code": {
                "coding": [
                    {
                        "system": "http://loinc.org",
                        "code": "55284-4",
                        "display": "Blood pressure",
                    }
                ],
                "text": "Blood pressure",
            },
            "components": [
                {
                    "valueQuantity": {"value": 120, "unit": "mm[Hg]"},
                    "code": {
                        "coding": [
                            {
                                "system": "http://loinc.org",
                                "code": "8480-6",
                                "display": "Systolic blood pressure",
                            }
                        ],
                        "text": "Systolic blood pressure",
                    },
                },
                {
                    "valueQuantity": {"value": 80, "unit": "mm[Hg]"},
                    "code": {
                        "coding": [
                            {
                                "system": "http://loinc.org",
                                "code": "8462-4",
                                "display": "Diastolic blood pressure",
                            }
                        ],
                        "text": "Diastolic blood pressure",
                    },
                },
            ],
        }
    )

    assert blood_pressure.components.systolic.valueQuantity.value == 120
    assert blood_pressure.systolic == 120
    assert blood_pressure.components.diastolic.valueQuantity.value == 80
    assert blood_pressure.diastolic == 80



from typing import Literal

from pydantic import PositiveInt, TypeAdapter

from fhir_slicing import BaseModel, Slice, SliceList, slice
from fhir_slicing.element_array import BaseElementArray
from fhir_slicing.extension import (
    BaseSimpleExtension,
    GeneralExtension,
)


def test_extension_model_get_url():
    class MyExtension(BaseSimpleExtension[Literal["http://example.com"], str]):
        valueString: str

    assert MyExtension.get_url() == "http://example.com"


def test_extension_array_from_extension_list():
    class MyExtensionA(BaseSimpleExtension[Literal["http://example.com/extension-a"], str]):
        valueString: str

    class MyExtensionB(BaseSimpleExtension[Literal["http://example.com/extension-b"], str]):
        valueString: str

    class ExtensionArray(BaseElementArray):
        a: SliceList[MyExtensionA] = slice(0, "*")
        b: Slice[MyExtensionB] = slice(1, 1)
        _: SliceList[GeneralExtension] = slice(0, "*")

    ext_list = [
        MyExtensionA(url="http://example.com/extension-a", valueString="a"),
        MyExtensionA(url="http://example.com/extension-a", valueString="a"),
        MyExtensionA(url="http://example.com/extension-a", valueString="a"),
        MyExtensionB(url="http://example.com/extension-b", valueString="b"),
        GeneralExtension.model_validate({"url": "http://example.com", "valueInteger": 3}),
    ]

    ext_array = ExtensionArray(ext_list)
    assert ext_array.a == ext_list[:3]
    assert ext_array.b == ext_list[3]
    assert list(ext_array) == ext_list


def test_extension_array_validator():
    class MyExtensionA(BaseSimpleExtension[Literal["http://example.com/extension-a"], str]):
        valueString: str

    class MyExtensionB(BaseSimpleExtension[Literal["http://example.com/extension-b"], str]):
        valueString: str

    class ExtensionArray(BaseElementArray):
        a: SliceList[MyExtensionA] = slice(0, "*")
        b: Slice[MyExtensionB] = slice(1, 1)
        _: SliceList[GeneralExtension] = slice(0, "*")

    ext_list = [
        {"url": "http://example.com", "valueInteger": 5},
        {"url": "http://example.com/extension-a", "valueString": "1"},
        {"url": "http://example.com/extension-a", "valueString": "2"},
        {"url": "http://example.com/extension-a", "valueString": "3"},
        {"url": "http://example.com/extension-b", "valueString": "4"},
    ]

    ext_array = TypeAdapter(ExtensionArray).validate_python(ext_list)

    assert ext_array.a == [
        MyExtensionA(url="http://example.com/extension-a", valueString="1"),
        MyExtensionA(url="http://example.com/extension-a", valueString="2"),
        MyExtensionA(url="http://example.com/extension-a", valueString="3"),
    ]

    assert ext_array.b == MyExtensionB(url="http://example.com/extension-b", valueString="4")

    ext_list_roundtrip = TypeAdapter(ExtensionArray).dump_python(ext_array, mode="python")
    assert ext_list_roundtrip == ext_list


def test_extension_array_ordering_roundtrip():
    class MyExtensionA(BaseSimpleExtension[Literal["http://example.com/extension-a"], str]):
        valueString: str

    class MyExtensionB(BaseSimpleExtension[Literal["http://example.com/extension-b"], str]):
        valueString: str

    class ExtensionArray(BaseElementArray):
        a: SliceList[MyExtensionA] = slice(0, "*")
        b: Slice[MyExtensionB] = slice(1, 1)

    ext_array = ExtensionArray(
        (
            MyExtensionA(url="http://example.com/extension-a", valueString="a"),
            MyExtensionA(url="http://example.com/extension-a", valueString="a"),
            MyExtensionA(url="http://example.com/extension-a", valueString="a"),
            MyExtensionB(url="http://example.com/extension-b", valueString="b"),
        )
    )

    ext_list = TypeAdapter(ExtensionArray).dump_python(ext_array)

    assert ext_list == [
        {"url": "http://example.com/extension-a", "valueString": "a"},
        {"url": "http://example.com/extension-a", "valueString": "a"},
        {"url": "http://example.com/extension-a", "valueString": "a"},
        {"url": "http://example.com/extension-b", "valueString": "b"},
    ]

    ext_array_roundtrip = TypeAdapter(ExtensionArray).validate_python(ext_list)

    assert ext_array_roundtrip == ext_array


def test_patient_use_case():
    class MultipleBirth(
        BaseSimpleExtension[Literal["http://hl7.org/fhir/StructureDefinition/patient-multipleBirth"], PositiveInt]
    ):
        valueInteger: PositiveInt

    class PatientExtensions(BaseElementArray):
        multiple_birth: Slice[MultipleBirth] = slice(1, 1)

    class PatientName(BaseModel):
        text: str
        given: list[str] | None = None
        family: str | None = None
        use: Literal["usual", "official", "temp", "nickname", "anounymous", "old", "maiden"] | None = None

    class Patient(BaseModel):
        extensions: PatientExtensions
        resourceType: Literal["Patient"] = "Patient"
        name: list[PatientName] | None = None

        @property
        def multiple_birth(self):
            return self.extensions.multiple_birth.valueInteger

        @multiple_birth.setter
        def set_multiple_birth(self, value: PositiveInt):
            self.extensions.multiple_birth.valueInteger = value

    patient = Patient.model_validate(
        {
            "resourceType": "Patient",
            "name": [
                {
                    "text": "John Doe",
                    "given": ["John"],
                    "family": "Doe",
                    "use": "official",
                },
            ],
            "extensions": [
                {
                    "url": "http://hl7.org/fhir/StructureDefinition/patient-multipleBirth",
                    "valueInteger": 3,
                }
            ],
        }
    )

    assert patient.extensions.multiple_birth.valueInteger == 3
    assert patient.multiple_birth == 3


def test_extension_array_with_subclassing():
    class MyExtensionA(BaseSimpleExtension[Literal["http://example.com/extension-a"], str]):
        valueString: str

    class MyExtensionB(BaseSimpleExtension[Literal["http://example.com/extension-b"], str]):
        valueString: str

    class ExtensionArray(BaseElementArray):
        a: SliceList[MyExtensionA] = slice(0, "*")
        _: SliceList[GeneralExtension] = slice(0, "*")

    class SubExtensionArray(ExtensionArray):
        b: Slice[MyExtensionB] = slice(1, 1)

    ext_list = [
        {"url": "http://example.com", "valueInteger": 5},
        {"url": "http://example.com/extension-a", "valueString": "1"},
        {"url": "http://example.com/extension-a", "valueString": "2"},
        {"url": "http://example.com/extension-a", "valueString": "3"},
        {"url": "http://example.com/extension-b", "valueString": "4"},
    ]

    ext_array = TypeAdapter(SubExtensionArray).validate_python(ext_list)

    assert ext_array.a == [
        MyExtensionA(url="http://example.com/extension-a", valueString="1"),
        MyExtensionA(url="http://example.com/extension-a", valueString="2"),
        MyExtensionA(url="http://example.com/extension-a", valueString="3"),
    ]

    assert ext_array.b == MyExtensionB(url="http://example.com/extension-b", valueString="4")

    ext_list_roundtrip = TypeAdapter(ExtensionArray).dump_python(ext_array, mode="python")
    assert ext_list_roundtrip == ext_list



from functools import partial
from typing import Any, Container, Iterable, List, Literal, TypeGuard, TypeVar, overload

from fhir_slicing.typing import ElementArray

TValueType = TypeVar("TValueType", covariant=True)


class BaseSlice[TValueType]:
    def __init__(self, min_items: int = 0, max_items: int | None = None):
        self.min_items = min_items
        self.max_items = max_items

    def __repr__(self):
        return f"{self.__class__.__name__}(min_items={self.min_items}, max_items={self.max_items})"

    def __set_name__(self, owner: ElementArray[Any], name: str):
        self.slice_name = name if name != "_" else "@default"
        self.filter_elements: partial[Iterable[TValueType]] = partial(
            owner.filter_elements_for_slice, slice_name=self.slice_name
        )
        self.is_element_part_of_slice: partial[TypeGuard[TValueType]] = partial(
            owner.is_element_part_of_slice, slice_name=self.slice_name
        )


class Slice[TValueType](BaseSlice[TValueType]):
    def __get__(self, obj: ElementArray[Any], objtype: type[Container] | None = None) -> TValueType:
        try:
            return next(iter(self.filter_elements(obj)))
        except StopIteration:
            raise ValueError(f"No value for slice '{self.slice_name}'.")

    def __set__(self, obj: ElementArray[TValueType], element: Any):
        for index, old_element in enumerate(obj):
            if self.is_element_part_of_slice(old_element):
                obj[index] = element
            return
        raise ValueError("Cannot set value on slice.")


class OptionalSlice[TValueType](BaseSlice[TValueType]):
    def __get__(self, obj: ElementArray[Any], objtype: type[Container] | None = None) -> TValueType | None:
        return next(iter(self.filter_elements(obj)), None)

    def __set__(self, obj: ElementArray[Any], element: TValueType):
        for index, old_element in enumerate(obj):
            if self.is_element_part_of_slice(old_element):
                obj[index] = element
                return


class SliceList[TValueType](BaseSlice[TValueType]):
    def __get__(self, obj: ElementArray[Any], objtype: type[Container] | None = None) -> List[TValueType]:
        return [*self.filter_elements(obj)]

    def __set__(self, obj: List, value: List):
        raise NotImplementedError("Cannot set value on slice list.")


@overload
def slice(min_items: Literal[0], max_items: Literal[1]) -> OptionalSlice: ...
@overload
def slice(min_items: Literal[1], max_items: Literal[1]) -> Slice: ...
@overload
def slice(min_items: Literal[0], max_items: Literal["*"]) -> SliceList: ...
@overload
def slice(min_items: Literal[1], max_items: Literal["*"]) -> SliceList: ...
@overload
def slice(min_items: int, max_items: int | Literal["*"]) -> Any: ...
def slice(min_items: int, max_items: int | Literal["*"]):
    match (min_items, max_items):
        case (0, 1):
            return OptionalSlice(min_items=0, max_items=1)
        case (1, 1):
            return Slice(min_items=1, max_items=1)
        case (0, "*"):
            return SliceList(min_items=0)
        case (1, "*"):
            return SliceList(min_items=1)
        case (int(), "*"):
            return SliceList(min_items=min_items)
        case (int(), int()):
            return SliceList(min_items=min_items)
        case _:
            raise ValueError(f"Invalid slice specification: {min_items}..{max_items}")


def slice_validator(elements: ElementArray, *, slice_name: str, slice_descriptor: BaseSlice):
    num_items = len(list(slice_descriptor.filter_elements(elements)))
    if slice_descriptor.max_items is not None and num_items > slice_descriptor.max_items:
        raise ValueError(f"Too many items in slice '{slice_name}'")
    elif num_items < slice_descriptor.min_items:
        raise ValueError(f"Not enough items in slice '{slice_name}'")
    return elements



from dataclasses import fields
from typing import (
    Self,
    cast,
)

from fhir_slicing.utils import get_value_from_literal

from .base import BaseModel


class BaseExtension[TUrl: str](BaseModel):
    url: TUrl


class GeneralExtension(BaseExtension):
    model_config = {"extra": "allow"}


class BaseSimpleExtension[TUrl: str, TValue](BaseExtension[TUrl]):
    url: TUrl

    @classmethod
    def get_url(cls) -> TUrl:
        return cast(TUrl, get_value_from_literal(cls.model_fields["url"].annotation))

    @property
    def value(self) -> TValue:
        value_field_name = next(field.name for field in fields(self) if field.name.startswith("value"))
        return getattr(self, value_field_name)

    @classmethod
    def from_value(cls, value: TValue) -> "Self":
        """Create an extension from a value"""
        value_field_name = next(field.name for field in fields(cls) if field.name.startswith("value"))
        return cls(url=cls.get_url(), **{value_field_name: value})


if __name__ == "__main__":
    pass



import types
from typing import (
    Annotated,
    Iterator,
    Literal,
    Type,
    TypeVar,
    Union,
    get_args,
    get_origin,
    overload,
)

from .slice import OptionalSlice, Slice, SliceList

T = TypeVar("T")


@overload
def get_source_type[T](annot, expect: Type[T]) -> Iterator[Type[T]]: ...
@overload
def get_source_type[T](annot, expect: None = None) -> Iterator[type]: ...
def get_source_type[T](annot, expect: Type[T] | None = None) -> Iterator[type | Type[T]]:
    """Extract the source type from a optional type or sequence type

    Example:
        get_source_type(Optional[str]) -> str
        get_source_type(List[str]) -> str
        get_source_type(str) -> str
        get_source_type(List[str]|None) -> str
        get_source_type(Annotated[str, "some annotation"]) -> str
        get_source_type(Annotated[List[str], "some annotation"]) -> str
        get_source_type(Annotated[List[str]|None, "some annotation"]) -> str
        get_source_type(Annotated[List[str|int]|None, "some annotateion"]) --> str, int
    """
    # If the annotation is a type, return the type
    if isinstance(annot, type):
        if expect is not None:
            if not issubclass(annot, expect):
                raise TypeError(f"Expected type to be a subclass of {expect}, got {annot}")
        yield annot
        return

    origin = get_origin(annot)
    if origin is None:
        yield annot

    elif origin is Annotated:
        yield from get_source_type(get_args(annot)[0], expect=expect)

    elif origin is list or origin is set:
        yield from get_source_type(get_args(annot)[0], expect=expect)

    elif origin in (SliceList, OptionalSlice, Slice):
        yield from get_source_type(get_args(annot)[0], expect=expect)

    # check for Union or UnionType
    elif origin is Union or isinstance(annot, types.UnionType):
        for arg in get_args(annot):
            if arg is not type(None):
                yield from get_source_type(arg, expect=expect)
    else:
        raise ValueError(f"Cannot determine source type from {annot}")


def get_value_from_literal(literal: type | None) -> int | str | None:
    """Get the value from a Literal type"""
    if get_origin(literal) is not Literal:
        return None
    return get_args(literal)[0]


# All FHIR Data Types
FHIRType = Literal[
    # Primitive Types
    "base64Binary",
    "boolean",
    "canonical",
    "code",
    "date",
    "dateTime",
    "decimal",
    "id",
    "instant",
    "integer",
    "integer64",
    "markdown",
    "oid",
    "positiveInt",
    "string",
    "time",
    "unsignedInt",
    "uri",
    "url",
    "uuid",
    # Complex Types
    "Address",
    "Age",
    "Annotation",
    "Attachment",
    "CodeableConcept",
    "CodeableReference",
    "Coding",
    "ContactPoint",
    "Count",
    "Distance",
    "Duration",
    "HumanName",
    "Identifier",
    "Money",
    "Period",
    "Quantity",
    "Range",
    "Ratio",
    "RatioRange",
    "Reference",
    "SampledData",
    "Signature",
    "Timing",
    # Metadata Types
    "ContactDetail",
    "DataRequirement",
    "Expression",
    "ExtendedContactDetail",
    "ParameterDefinition",
    "RelatedArtifact",
    "TriggerDefinition",
    "UsageContext",
    "Availability",
    # Special Types
    "Dosage",
    "Element",
    "Extension",
    "Meta",
    "Narrative",
]



import logging
from types import UnionType
from typing import Mapping, Type

from pydantic import GetCoreSchemaHandler
from pydantic_core import CoreSchema, core_schema
from typing_extensions import Any

from fhir_slicing.utils import get_source_type, get_value_from_literal

LOGGER = logging.getLogger(__name__)


def check_if_type_is_union_type(type_: Any) -> bool:
    return isinstance(type_, UnionType)


def get_schema_for_slices(handler: GetCoreSchemaHandler, slice_annotations: dict[str, Type]):
    """Generate a schema for each slice.

    Args:
        handler (GetCoreSchemaHandler): The handler to generate the schema

    Yields:
        tuple[str, CoreSchema]: The name of the slice and the schema
    """
    for slice_name, slice_type in slice_annotations.items():
        source_types = list(get_source_type(slice_type))
        match len(source_types):
            case 0:
                raise TypeError(f"Expected a source type, got none for {slice_name}")
            case 1:
                yield slice_name, handler(source_types[0])
            case _:
                yield slice_name, core_schema.union_schema([handler(source_type) for source_type in source_types])


def get_value_to_slice_name_map[TSliceType](
    slice_annotations: dict[str, Type], slice_type: Type[TSliceType] | None = None
) -> dict[str, str]:
    map = {}
    for slice_name, slice_annotation in slice_annotations.items():
        source_types = list(get_source_type(slice_annotation, expect=slice_type))
        if len(source_types) > 1:
            raise NotImplementedError(f"Slice {slice_name} has multiple source types. This is not supported")
        slice_value = get_value_from_literal(source_types[0])
        if slice_value is None:
            raise TypeError(f"Slice {slice_name} has no value")
        if slice_value in map:
            raise TypeError(f"Slice {slice_name} has duplicate value")
        map[slice_value] = slice_name
    return map


def get_slice_union_with_value_discriminator_schema(
    source_type: UnionType,
    path: str,
    handler: GetCoreSchemaHandler,
    slice_annotations: dict[str, Type],
) -> CoreSchema:
    value_to_slice_name = get_value_to_slice_name_map(slice_annotations)

    def discriminator(value: Any):
        if isinstance(value, Mapping):
            return value_to_slice_name.get(value.get(path, {}), None)
        LOGGER.warning(
            "⚠️Encountered slice type which is not a mapping. Make sure to implement the Mapping protocol on all slice types."
        )
        return None

    choices: dict[str, CoreSchema] = dict(get_schema_for_slices(handler, slice_annotations=slice_annotations))

    return core_schema.tagged_union_schema(choices=choices, discriminator=discriminator)


def get_slice_union_schema(
    source_type: UnionType, handler: GetCoreSchemaHandler, slice_annotations: dict[str, Type]
) -> CoreSchema:
    # type hint is necessary for union_schema input
    choices: list[CoreSchema | tuple[CoreSchema, str]] = [
        (slice_schema, slice_name)
        for slice_name, slice_schema in get_schema_for_slices(handler, slice_annotations=slice_annotations)
    ]

    return core_schema.union_schema(choices=choices)
